/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/swerve/impl/SwerveDrivetrainImpl.hpp"
#include "ctre/phoenix6/swerve/utility/PhoenixPIDController.hpp"

namespace ctre {
namespace phoenix6 {
namespace swerve {
namespace requests {

/**
 * \brief The reference for "forward" is sometimes different if you're talking
 * about field relative. This addresses which forward to use.
 */
enum class ForwardReferenceValue {
    /**
     * \brief This forward reference makes it so "forward" (positive X) is always towards the red alliance.
     * This is important in situations such as path following where positive X is always towards the
     * red alliance wall, regardless of where the operator physically are located.
     */
    RedAlliance = 0,
    /**
     * \brief This forward references makes it so "forward" (positive X) is determined from the operator's
     * perspective. This is important for most teleop driven field-centric requests, where positive
     * X really means to drive away from the operator.
     *
     * Important: Users must specify the OperatorPerspective in the SwerveDrivetrain object
     */
    OperatorPerspective = 1,
};

/**
 * \brief Container for all the Swerve Requests. Use this to find all applicable swerve
 * drive requests.
 *
 * This is also an interface common to all swerve drive control requests that
 * allow the request to calculate the state to apply to the modules.
 */
class SwerveRequest {
public:
    using ControlParameters = impl::SwerveDrivetrainImpl::ControlParameters;

    virtual ~SwerveRequest() = default;

    /**
     * \brief Applies this swerve request to the given modules.
     * This is typically called by the SwerveDrivetrain.
     *
     * \param parameters Parameters the control request needs to calculate the module state
     * \param modulesToApply Modules to which the control request is applied
     * \returns Status code of sending the request
     */
    virtual ctre::phoenix::StatusCode Apply(ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) = 0;
};

/**
 * \brief Does nothing to the swerve module state. This is the default state of a newly
 * created swerve drive mechanism.
 */
class Idle : public SwerveRequest {
public:
    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &) override
    {
        return ctre::phoenix::StatusCode::OK;
    }
};

/**
 * \brief Sets the swerve drive module states to point inward on the
 * robot in an "X" fashion, creating a natural brake which will
 * oppose any motion.
 */
class SwerveDriveBrake : public SwerveRequest {
public:
    /**
     * The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            impl::SwerveModuleState const state{0_mps, parameters.swervePositions[i].Angle()};
            modulesToApply[i]->Apply(state, DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    SwerveDriveBrake &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    SwerveDriveBrake &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }
};

/**
 * \brief Drives the swerve drivetrain in a field-centric manner.
 *
 * When users use this request, they specify the direction the robot should
 * travel oriented against the field, and the rate at which their robot should
 * rotate about the center of the robot.
 *
 * An example scenario is that the robot is oriented to the east, the
 * VelocityX is +5 m/s, VelocityY is 0 m/s, and RotationRate is 0.5 rad/s.
 * In this scenario, the robot would drive northward at 5 m/s and turn
 * counterclockwise at 0.5 rad/s.
 */
class FieldCentric : public SwerveRequest {
public:
    /**
     * \brief The velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     */
    units::meters_per_second_t VelocityX = 0_mps;
    /**
     * \brief The velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     */
    units::meters_per_second_t VelocityY = 0_mps;
    /**
     * \brief The angular rate to rotate at.
     * Angular rate is defined as counterclockwise positive,
     * so this determines how fast to turn counterclockwise.
     */
    units::radians_per_second_t RotationalRate = 0_rad_per_s;
    /**
     * \brief The allowable deadband of the request.
     */
    units::meters_per_second_t Deadband = 0_mps;
    /**
     * \brief The rotational deadband of the request.
     */
    units::radians_per_second_t RotationalDeadband = 0_rad_per_s;
    /**
     * \brief The center of rotation the robot should rotate around.
     * This is (0,0) by default, which will rotate around the center of the robot.
     */
    Translation2d CenterOfRotation{};

    /**
     * \brief The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * \brief The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;
    /**
     * \brief Whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     */
    bool DesaturateWheelSpeeds = true;

    /**
     * \brief The perspective to use when determining which direction is forward.
     */
    ForwardReferenceValue ForwardReference = ForwardReferenceValue::OperatorPerspective;

public:
    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        auto toApplyX = VelocityX;
        auto toApplyY = VelocityY;
        if (ForwardReference == ForwardReferenceValue::OperatorPerspective) {
            /* If we're operator perspective, modify the X/Y translation by the angle */
            Translation2d tmp{toApplyX * 1_s, toApplyY * 1_s};
            tmp = tmp.RotateBy(parameters.operatorForwardDirection);
            toApplyX = tmp.X() / 1_s;
            toApplyY = tmp.Y() / 1_s;
        }
        auto toApplyOmega = RotationalRate;
        if (units::math::hypot(toApplyX, toApplyY) < Deadband) {
            toApplyX = 0_mps;
            toApplyY = 0_mps;
        }
        if (units::math::abs(toApplyOmega) < RotationalDeadband) {
            toApplyOmega = 0_rad_per_s;
        }

        auto const speeds = impl::ChassisSpeeds::Discretize(impl::ChassisSpeeds::FromFieldRelativeSpeeds({toApplyX, toApplyY, toApplyOmega},
                parameters.currentPose.Rotation()), parameters.updatePeriod);

        auto states = parameters.kinematics->ToSwerveModuleStates(speeds, CenterOfRotation);
        if (DesaturateWheelSpeeds && parameters.kMaxSpeed > 0_mps) {
            impl::SwerveDriveKinematics::DesaturateWheelSpeeds(&states, parameters.kMaxSpeed);
        }

        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(states[i], DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     *
     * \param velocityX Velocity in the X direction
     * \returns this request
     */
    FieldCentric &WithVelocityX(units::meters_per_second_t velocityX)
    {
        this->VelocityX = velocityX;
        return *this;
    }

    /**
     * \brief Sets the velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     *
     * \param velocityY Velocity in the Y direction
     * \returns this request
     */
    FieldCentric &WithVelocityY(units::meters_per_second_t velocityY)
    {
        this->VelocityY = velocityY;
        return *this;
    }

    /**
     * \brief The angular rate to rotate at.
     * Angular rate is defined as counterclockwise positive,
     * so this determines how fast to turn counterclockwise.
     *
     * \param rotationalRate Angular rate to rotate at
     * \returns this request
     */
    FieldCentric &WithRotationalRate(units::radians_per_second_t rotationalRate)
    {
        this->RotationalRate = rotationalRate;
        return *this;
    }

    /**
     * \brief Sets the allowable deadband of the request.
     *
     * \param deadband Allowable deadband of the request
     * \returns this request
     */
    FieldCentric &WithDeadband(units::meters_per_second_t deadband)
    {
        this->Deadband = deadband;
        return *this;
    }
    /**
     * \brief Sets the rotational deadband of the request.
     *
     * \param rotationalDeadband Rotational deadband of the request
     * \returns this request
     */
    FieldCentric &WithRotationalDeadband(units::radians_per_second_t rotationalDeadband)
    {
        this->RotationalDeadband = rotationalDeadband;
        return *this;
    }
    /**
     * \brief Sets the center of rotation of the request
     *
     * \param centerOfRotation The center of rotation the robot should rotate around.
     * \returns this request
     */
    FieldCentric &WithCenterOfRotation(Translation2d centerOfRotation)
    {
        this->CenterOfRotation = centerOfRotation;
        return *this;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    FieldCentric &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    FieldCentric &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }

    /**
     * \brief Sets whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     *
     * \param desaturateWheelSpeeds Whether to desaturate wheel speeds
     * \returns this request
     */
    FieldCentric &WithDesaturateWheelSpeeds(bool desaturateWheelSpeeds)
    {
        this->DesaturateWheelSpeeds = desaturateWheelSpeeds;
        return *this;
    }

    /**
     * \brief Sets the perspective to use when determining which direction is forward.
     *
     * \param desaturateWheelSpeeds The perspective to use when determining which direction is forward
     * \returns this request
     */
    FieldCentric &WithForwardReference(ForwardReferenceValue forwardReference)
    {
        this->ForwardReference = forwardReference;
        return *this;
    }
};

/**
 * \brief Drives the swerve drivetrain in a field-centric manner, maintaining a
 * specified heading angle to ensure the robot is facing the desired direction
 *
 * When users use this request, they specify the direction the robot should
 * travel oriented against the field, and the direction the robot should be facing.
 *
 * An example scenario is that the robot is oriented to the east, the VelocityX
 * is +5 m/s, VelocityY is 0 m/s, and TargetDirection is 180 degrees.
 * In this scenario, the robot would drive northward at 5 m/s and turn clockwise
 * to a target of 180 degrees.
 *
 * This control request is especially useful for autonomous control, where the
 * robot should be facing a changing direction throughout the motion.
 */
class FieldCentricFacingAngle : public SwerveRequest {
public:
    /**
     * \brief The velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     */
    units::meters_per_second_t VelocityX = 0_mps;
    /**
     * \brief The velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     */
    units::meters_per_second_t VelocityY = 0_mps;
    /**
     * \brief The desired direction to face.
     * 0 Degrees is defined as in the direction of the X axis.
     * As a result, a TargetDirection of 90 degrees will point along
     * the Y axis, or to the left.
     */
    Rotation2d TargetDirection{};

    /**
     * \brief The allowable deadband of the request.
     */
    units::meters_per_second_t Deadband = 0_mps;
    /**
     * \brief The rotational deadband of the request.
     */
    units::radians_per_second_t RotationalDeadband = 0_rad_per_s;
    /**
     * \brief The center of rotation the robot should rotate around.
     * This is (0,0) by default, which will rotate around the center of the robot.
     */
    Translation2d CenterOfRotation{};

    /**
     * \brief The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * \brief The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;
    /**
     * \brief Whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     */
    bool DesaturateWheelSpeeds = true;

    /**
     * \brief The perspective to use when determining which direction is forward.
     */
    ForwardReferenceValue ForwardReference = ForwardReferenceValue::OperatorPerspective;

    /**
     * \brief The PID controller used to maintain the desired heading.
     * Users can specify the PID gains to change how aggressively to maintain
     * heading.
     *
     * This PID controller operates on heading radians and outputs a target
     * rotational rate in radians per second.
     */
    PhoenixPIDController HeadingController{0, 0, 0};

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        auto toApplyX = VelocityX;
        auto toApplyY = VelocityY;
        Rotation2d angleToFace = TargetDirection;
        if (ForwardReference == ForwardReferenceValue::OperatorPerspective) {
            /* If we're operator perspective, modify the X/Y translation by the angle */
            Translation2d tmp{toApplyX * 1_s, toApplyY * 1_s};
            tmp = tmp.RotateBy(parameters.operatorForwardDirection);
            toApplyX = tmp.X() / 1_s;
            toApplyY = tmp.Y() / 1_s;
            /* And rotate the direction we want to face by the angle */
            angleToFace = angleToFace.RotateBy(parameters.operatorForwardDirection);
        }

        units::radians_per_second_t toApplyOmega{HeadingController.Calculate(parameters.currentPose.Rotation().Radians().value(),
                angleToFace.Radians().value(), parameters.timestamp)};
        if (units::math::hypot(toApplyX, toApplyY) < Deadband) {
            toApplyX = 0_mps;
            toApplyY = 0_mps;
        }
        if (units::math::abs(toApplyOmega) < RotationalDeadband) {
            toApplyOmega = 0_rad_per_s;
        }

        auto const speeds = impl::ChassisSpeeds::Discretize(impl::ChassisSpeeds::FromFieldRelativeSpeeds({toApplyX, toApplyY, toApplyOmega},
                parameters.currentPose.Rotation()), parameters.updatePeriod);

        auto states = parameters.kinematics->ToSwerveModuleStates(speeds, CenterOfRotation);
        if (DesaturateWheelSpeeds && parameters.kMaxSpeed > 0_mps) {
            impl::SwerveDriveKinematics::DesaturateWheelSpeeds(&states, parameters.kMaxSpeed);
        }

        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(states[i], DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     *
     * \param velocityX Velocity in the X direction
     * \returns this request
     */
    FieldCentricFacingAngle &WithVelocityX(units::meters_per_second_t velocityX)
    {
        this->VelocityX = velocityX;
        return *this;
    }

    /**
     * \brief Sets the velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     *
     * \param velocityY Velocity in the Y direction
     * \returns this request
     */
    FieldCentricFacingAngle &WithVelocityY(units::meters_per_second_t velocityY)
    {
        this->VelocityY = velocityY;
        return *this;
    }

    /**
     * \brief Sets the desired direction to face.
     * 0 Degrees is defined as in the direction of the X axis.
     * As a result, a TargetDirection of 90 degrees will point along
     * the Y axis, or to the left.
     *
     * \param targetDirection Desired direction to face
     * \returns this request
     */
    FieldCentricFacingAngle &WithTargetDirection(Rotation2d targetDirection)
    {
        this->TargetDirection = targetDirection;
        return *this;
    }

    /**
     * \brief Sets the allowable deadband of the request.
     *
     * \param deadband Allowable deadband of the request
     * \returns this request
     */
    FieldCentricFacingAngle &WithDeadband(units::meters_per_second_t deadband)
    {
        this->Deadband = deadband;
        return *this;
    }
    /**
     * \brief Sets the rotational deadband of the request.
     *
     * \param rotationalDeadband Rotational deadband of the request
     * \returns this request
     */
    FieldCentricFacingAngle &WithRotationalDeadband(units::radians_per_second_t rotationalDeadband)
    {
        this->RotationalDeadband = rotationalDeadband;
        return *this;
    }
    /**
     * \brief Sets the center of rotation of the request
     *
     * \param centerOfRotation The center of rotation the robot should rotate around.
     * \returns this request
     */
    FieldCentricFacingAngle &WithCenterOfRotation(Translation2d centerOfRotation)
    {
        this->CenterOfRotation = centerOfRotation;
        return *this;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    FieldCentricFacingAngle &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    FieldCentricFacingAngle &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }

    /**
     * \brief Sets whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     *
     * \param desaturateWheelSpeeds Whether to desaturate wheel speeds
     * \returns this request
     */
    FieldCentricFacingAngle &WithDesaturateWheelSpeeds(bool desaturateWheelSpeeds)
    {
        this->DesaturateWheelSpeeds = desaturateWheelSpeeds;
        return *this;
    }

    /**
     * \brief Sets the perspective to use when determining which direction is forward.
     *
     * \param desaturateWheelSpeeds The perspective to use when determining which direction is forward
     * \returns this request
     */
    FieldCentricFacingAngle &WithForwardReference(ForwardReferenceValue forwardReference)
    {
        this->ForwardReference = forwardReference;
        return *this;
    }
};

/**
 * \brief Sets the swerve drive modules to point to a specified direction.
 */
class PointWheelsAt : public SwerveRequest {
public:
    /**
     * \brief The direction to point the modules toward.
     * This direction will still be optimized to what the module was previously at.
     */
    Rotation2d ModuleDirection{};
    /**
     * \brief The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * \brief The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            impl::SwerveModuleState const state{0_mps, ModuleDirection};
            modulesToApply[i]->Apply(state, DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the direction to point the modules toward.
     * This direction is still optimized to what the module was previously at.
     *
     * \param moduleDirection Direction to point the modules toward
     * \returns this request
     */
    PointWheelsAt &WithModuleDirection(Rotation2d moduleDirection)
    {
        this->ModuleDirection = moduleDirection;
        return *this;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    PointWheelsAt &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    PointWheelsAt &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }
};

/**
 * \brief Drives the swerve drivetrain in a robot-centric manner.
 *
 * When users use this request, they specify the direction the robot should
 * travel oriented against the robot itself, and the rate at which their
 * robot should rotate about the center of the robot.
 *
 * An example scenario is that the robot is oriented to the east, the
 * VelocityX is +5 m/s, VelocityY is 0 m/s, and RotationRate is 0.5 rad/s.
 * In this scenario, the robot would drive eastward at 5 m/s and turn
 * counterclockwise at 0.5 rad/s.
 */
class RobotCentric : public SwerveRequest {
public:
    /**
     * \brief The velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     */
    units::meters_per_second_t VelocityX = 0_mps;
    /**
     * \brief The velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     */
    units::meters_per_second_t VelocityY = 0_mps;
    /**
     * \brief The angular rate to rotate at.
     * Angular rate is defined as counterclockwise positive,
     * so this determines how fast to turn counterclockwise.
     */
    units::radians_per_second_t RotationalRate = 0_rad_per_s;

    /**
     * \brief The allowable deadband of the request.
     */
    units::meters_per_second_t Deadband = 0_mps;
    /**
     * \brief The rotational deadband of the request.
     */
    units::radians_per_second_t RotationalDeadband = 0_rad_per_s;
    /**
     * \brief The center of rotation the robot should rotate around.
     * This is (0,0) by default, which will rotate around the center of the robot.
     */
    Translation2d CenterOfRotation{};

    /**
     * \brief The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * \brief The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;
    /**
     * \brief Whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     */
    bool DesaturateWheelSpeeds = true;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        auto toApplyX = VelocityX;
        auto toApplyY = VelocityY;
        auto toApplyOmega = RotationalRate;
        if (units::math::hypot(toApplyX, toApplyY) < Deadband) {
            toApplyX = 0_mps;
            toApplyY = 0_mps;
        }
        if (units::math::abs(toApplyOmega) < RotationalDeadband) {
            toApplyOmega = 0_rad_per_s;
        }
        impl::ChassisSpeeds const speeds{toApplyX, toApplyY, toApplyOmega};

        auto states = parameters.kinematics->ToSwerveModuleStates(speeds, CenterOfRotation);
        if (DesaturateWheelSpeeds && parameters.kMaxSpeed > 0_mps) {
            impl::SwerveDriveKinematics::DesaturateWheelSpeeds(&states, parameters.kMaxSpeed);
        }

        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(states[i], DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the velocity in the X direction.
     * X is defined as forward according to WPILib convention,
     * so this determines how fast to travel forward.
     *
     * \param velocityX Velocity in the X direction
     * \returns this request
     */
    RobotCentric &WithVelocityX(units::meters_per_second_t velocityX)
    {
        this->VelocityX = velocityX;
        return *this;
    }

    /**
     * \brief Sets the velocity in the Y direction.
     * Y is defined as to the left according to WPILib convention,
     * so this determines how fast to travel to the left.
     *
     * \param velocityY Velocity in the Y direction
     * \returns this request
     */
    RobotCentric &WithVelocityY(units::meters_per_second_t velocityY)
    {
        this->VelocityY = velocityY;
        return *this;
    }

    /**
     * \brief The angular rate to rotate at.
     * Angular rate is defined as counterclockwise positive,
     * so this determines how fast to turn counterclockwise.
     *
     * \param rotationalRate Angular rate to rotate at
     * \returns this request
     */
    RobotCentric &WithRotationalRate(units::radians_per_second_t rotationalRate)
    {
        this->RotationalRate = rotationalRate;
        return *this;
    }

    /**
     * \brief Sets the allowable deadband of the request.
     *
     * \param deadband Allowable deadband of the request
     * \returns this request
     */
    RobotCentric &WithDeadband(units::meters_per_second_t deadband)
    {
        this->Deadband = deadband;
        return *this;
    }
    /**
     * \brief Sets the rotational deadband of the request.
     *
     * \param rotationalDeadband Rotational deadband of the request
     * \returns this request
     */
    RobotCentric &WithRotationalDeadband(units::radians_per_second_t rotationalDeadband)
    {
        this->RotationalDeadband = rotationalDeadband;
        return *this;
    }
    /**
     * \brief Sets the center of rotation of the request
     *
     * \param centerOfRotation The center of rotation the robot should rotate around.
     * \returns this request
     */
    RobotCentric &WithCenterOfRotation(Translation2d centerOfRotation)
    {
        this->CenterOfRotation = centerOfRotation;
        return *this;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    RobotCentric &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    RobotCentric &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }

    /**
     * \brief Sets whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     *
     * \param desaturateWheelSpeeds Whether to desaturate wheel speeds
     * \returns this request
     */
    RobotCentric &WithDesaturateWheelSpeeds(bool desaturateWheelSpeeds)
    {
        this->DesaturateWheelSpeeds = desaturateWheelSpeeds;
        return *this;
    }
};

/**
 * \brief Accepts a generic ChassisSpeeds to apply to the drivetrain.
 */
class ApplyChassisSpeeds : public SwerveRequest {
public:
    /**
     * \brief The chassis speeds to apply to the drivetrain.
     */
    impl::ChassisSpeeds Speeds{};
    /**
     * \brief The center of rotation to rotate around.
     */
    Translation2d CenterOfRotation{};
    /**
     * \brief The type of control request to use for the drive motor.
     */
    impl::SwerveModuleImpl::DriveRequestType DriveRequestType = impl::SwerveModuleImpl::DriveRequestType::OpenLoopVoltage;
    /**
     * \brief The type of control request to use for the steer motor.
     */
    impl::SwerveModuleImpl::SteerRequestType SteerRequestType = impl::SwerveModuleImpl::SteerRequestType::MotionMagicExpo;
    /**
     * \brief Whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     */
    bool DesaturateWheelSpeeds = true;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        auto states = parameters.kinematics->ToSwerveModuleStates(Speeds, CenterOfRotation);
        if (DesaturateWheelSpeeds && parameters.kMaxSpeed > 0_mps) {
            impl::SwerveDriveKinematics::DesaturateWheelSpeeds(&states, parameters.kMaxSpeed);
        }

        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(states[i], DriveRequestType, SteerRequestType);
        }

        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the chassis speeds to apply to the drivetrain.
     *
     * \param speeds Chassis speeds to apply to the drivetrain
     * \returns this request
     */
    ApplyChassisSpeeds &WithSpeeds(impl::ChassisSpeeds speeds)
    {
        this->Speeds = speeds;
        return *this;
    }
    /**
     * \brief Sets the center of rotation to rotate around.
     *
     * \param centerOfRotation Center of rotation to rotate around
     * \returns this request
     */
    ApplyChassisSpeeds &WithCenterOfRotation(Translation2d centerOfRotation)
    {
        this->CenterOfRotation = centerOfRotation;
        return *this;
    }

    /**
     * \brief Sets the type of control request to use for the drive motor.
     *
     * \param driveRequestType The type of control request to use for the drive motor
     * \returns this request
     */
    ApplyChassisSpeeds &WithDriveRequestType(impl::SwerveModuleImpl::DriveRequestType driveRequestType)
    {
        this->DriveRequestType = driveRequestType;
        return *this;
    }
    /**
     * \brief Sets the type of control request to use for the steer motor.
     *
     * \param steerRequestType The type of control request to use for the steer motor
     * \returns this request
     */
    ApplyChassisSpeeds &WithSteerRequestType(impl::SwerveModuleImpl::SteerRequestType steerRequestType)
    {
        this->SteerRequestType = steerRequestType;
        return *this;
    }

    /**
     * \brief Sets whether to desaturate wheel speeds before applying.
     * For more information, see the documentation of impl#SwerveDriveKinematics#DesaturateWheelSpeeds.
     *
     * \param desaturateWheelSpeeds Whether to desaturate wheel speeds
     * \returns this request
     */
    ApplyChassisSpeeds &WithDesaturateWheelSpeeds(bool desaturateWheelSpeeds)
    {
        this->DesaturateWheelSpeeds = desaturateWheelSpeeds;
        return *this;
    }
};

/**
 * \brief SysId-specific SwerveRequest to characterize the translational
 * characteristics of a swerve drivetrain.
 */
class SysIdSwerveTranslation : public SwerveRequest {
public:
    /**
     * \brief Voltage to apply to drive wheels.
     */
    units::volt_t VoltsToApply = 0_V;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(controls::VoltageOut{VoltsToApply}, controls::PositionVoltage{0_tr});
        }
        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Sets the voltage to apply to drive wheels.
     *
     * \param volts Voltage to apply
     * \returns this request
     */
    SysIdSwerveTranslation &WithVolts(units::volt_t volts)
    {
        this->VoltsToApply = volts;
        return *this;
    }
};

/**
 * \brief SysId-specific SwerveRequest to characterize the rotational
 * characteristics of a swerve drivetrain.
 */
class SysIdSwerveRotation : public SwerveRequest {
public:
    /**
     * \brief Voltage to apply to drive wheels.
     */
    units::volt_t VoltsToApply = 0_V;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(controls::VoltageOut{VoltsToApply},
                                    controls::PositionVoltage{(parameters.swervePositions[i].Angle() + Rotation2d{90_deg}).Radians()});
        }
        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Update the voltage to apply to the drive wheels.
     *
     * \param volts Voltage to apply
     * \returns this request
     */
    SysIdSwerveRotation &WithVolts(units::volt_t volts)
    {
        this->VoltsToApply = volts;
        return *this;
    }
};

/**
 * \brief SysId-specific SwerveRequest to characterize the steer module
 * characteristics of a swerve drivetrain.
 */
class SysIdSwerveSteerGains : public SwerveRequest {
public:
    /**
     * \brief Voltage to apply to drive wheels.
     */
    units::volt_t VoltsToApply = 0_V;

    ctre::phoenix::StatusCode Apply(SwerveRequest::ControlParameters const &parameters, std::vector<std::unique_ptr<impl::SwerveModuleImpl>> const &modulesToApply) override
    {
        for (size_t i = 0; i < modulesToApply.size(); ++i) {
            modulesToApply[i]->Apply(controls::VoltageOut{0_V}, controls::VoltageOut{VoltsToApply});
        }
        return ctre::phoenix::StatusCode::OK;
    }

    /**
     * \brief Update the voltage to apply to the drive wheels.
     *
     * \param volts Voltage to apply
     * \returns this request
     */
    SysIdSwerveSteerGains &WithVolts(units::volt_t volts)
    {
        this->VoltsToApply = volts;
        return *this;
    }
};

}
}
}
}
