/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/swerve/SwerveModuleConstants.hpp"
#include "ctre/phoenix6/swerve/impl/SwerveDriveKinematics.hpp"
#include <array>

namespace ctre {
namespace phoenix6 {
namespace swerve {
namespace impl {

/**
 * \brief Swerve Module class that encapsulates a swerve module powered by
 * CTR Electronics devices.
 *
 * This class handles the hardware devices but does not configure them for
 * swerve module operation using the Phoenix 6 API. Users should create a
 * high-level SwerveModule instead of using this directly.
 */
class SwerveModuleImpl {
public:
    /**
     * \brief All possible control requests for the module steer motor.
     */
    enum class SteerRequestType {
        /**
         * \brief Control the drive motor using a Motion Magic® Expo request.
         * The control output type is determined by SwerveModuleConstants#SteerMotorClosedLoopOutput.
         */
        MotionMagicExpo = 0,
        /**
         * \brief Control the drive motor using an unprofiled position request.
         * The control output type is determined by SwerveModuleConstants#SteerMotorClosedLoopOutput.
         */
        Position = 1,
    };

    /**
     * \brief All possible control requests for the module drive motor.
     */
    enum class DriveRequestType {
        /**
         * \brief Control the drive motor using an open-loop voltage request.
         */
        OpenLoopVoltage = 0,
        /**
         * \brief Control the drive motor using a velocity closed-loop request.
         * The control output type is determined by SwerveModuleConstants#DriveMotorClosedLoopOutput.
         */
        Velocity = 1,
    };

private:
    hardware::core::CoreTalonFX _driveMotor;
    hardware::core::CoreTalonFX _steerMotor;
    hardware::core::CoreCANcoder _cancoder;

    StatusSignal<units::turn_t> _drivePosition;
    StatusSignal<units::turns_per_second_t> _driveVelocity;
    StatusSignal<units::turn_t> _steerPosition;
    StatusSignal<units::turns_per_second_t> _steerVelocity;

    ClosedLoopOutputType _driveClosedLoopOutput;
    ClosedLoopOutputType _steerClosedLoopOutput;

    using turns_per_meter = units::compound_unit<units::turns, units::inverse<units::meters>>;
    using turns_per_meter_t = units::unit_t<turns_per_meter>;

    turns_per_meter_t kDriveRotationsPerMeter;
    units::scalar_t kCouplingRatioDriveRotorToCANcoder;
    units::meters_per_second_t kSpeedAt12Volts;

    bool _isOnCANFD;

    SwerveModulePosition _internalState;
    SwerveModuleState _targetState;

public:
    /**
     * \brief Construct a SwerveModuleImpl with the specified constants.
     *
     * \param constants   Constants used to construct the module
     * \param canbusName  The name of the CAN bus this module is on
     */
    SwerveModuleImpl(SwerveModuleConstants const &constants, std::string_view canbusName);

    /**
     * \brief Applies the desired SwerveModuleState to this module.
     *
     * \param state            Speed and direction the module should target
     * \param driveRequestType The #DriveRequestType to apply
     * \param steerRequestType The #SteerRequestType to apply; defaults to SteerRequestType#MotionMagicExpo
     */
    void Apply(SwerveModuleState const &state, DriveRequestType driveRequestType, SteerRequestType steerRequestType = SteerRequestType::MotionMagicExpo);

    /**
     * \brief Controls this module using the specified drive and steer control requests.
     *
     * This is intended only to be used for characterization of the robot; do not use this for normal use.
     *
     * \param driveRequest The control request to apply to the drive motor
     * \param steerRequest The control request to apply to the steer motor
     */
    template <typename DriveReq, typename SteerReq>
    void Apply(DriveReq &&driveRequest, SteerReq &&steerRequest)
    {
        _driveMotor.SetControl(driveRequest.WithUpdateFreqHz(0_Hz));
        _steerMotor.SetControl(steerRequest.WithUpdateFreqHz(0_Hz));
    }

    /**
     * \brief Configures the neutral mode to use for the module's drive motor.
     *
     * \param neutralMode The drive motor neutral mode
     * \param timeoutSeconds Maximum amount of time to wait when performing configuration
     * \returns Status code response of the request
     */
    ctre::phoenix::StatusCode ConfigNeutralMode(signals::NeutralModeValue neutralMode, units::second_t timeoutSeconds = 0.100_s);

    /**
     * \brief Gets the state of this module and passes it back as a
     * SwerveModulePosition object with latency compensated values.
     *
     * \param refresh True if the signals should be refreshed
     * \returns SwerveModulePosition containing this module's state.
     */
    SwerveModulePosition GetPosition(bool refresh);

    /**
     * \brief Gets the last cached swerve module position.
     * This differs from #GetPosition in that it will not
     * perform any latency compensation or refresh the signals.
     *
     * \returns Last cached SwerveModulePosition
     */
    SwerveModulePosition GetCachedPosition() const { return _internalState; }

    /**
     * \brief Get the current state of the module.
     *
     * This is typically used for telemetry, as the SwerveModulePosition
     * is used for odometry.
     *
     * \returns Current state of the module
     */
    SwerveModuleState GetCurrentState() const
    {
        return SwerveModuleState{_driveVelocity.GetValue() / kDriveRotationsPerMeter, {_steerPosition.GetValue()}};
    }

    /**
     * \brief Get the target state of the module.
     *
     * This is typically used for telemetry.
     *
     * \returns Target state of the module
     */
    SwerveModuleState GetTargetState() const { return _targetState; }

    /**
     * Resets this module's drive motor position to 0 rotations.
     */
    void ResetPosition()
    {
        /* Only touch drive pos, not steer */
        _driveMotor.SetPosition(0_tr);
    }

    /**
     * Gets this module's Drive Motor TalonFX reference.
     *
     * This should be used only to access signals and change configurations that the
     * swerve drivetrain does not configure itself.
     *
     * \returns This module's Drive Motor reference
     */
    hardware::core::CoreTalonFX &GetDriveMotor() { return _driveMotor; }

    /**
     * Gets this module's Steer Motor TalonFX reference.
     *
     * This should be used only to access signals and change configurations that the
     * swerve drivetrain does not configure itself.
     *
     * \returns This module's Steer Motor reference
     */
    hardware::core::CoreTalonFX &GetSteerMotor() { return _steerMotor; }

    /**
     * Gets this module's CANcoder reference.
     *
     * This should be used only to access signals and change configurations that the
     * swerve drivetrain does not configure itself.
     *
     * \returns This module's CANcoder reference
     */
    hardware::core::CoreCANcoder &GetCANcoder() { return _cancoder; }

private:
    friend class SwerveDrivetrainImpl;

    std::array<BaseStatusSignal *, 4> GetSignals()
    {
        return std::array<BaseStatusSignal *, 4>{&_drivePosition, &_driveVelocity, &_steerPosition, &_steerVelocity};
    }
};

}
}
}
}
